/*
   +----------------------------------------------------------------------+
   | PHP version 4.0                                                      |
   +----------------------------------------------------------------------+
   | Copyright (c) 1997-2001 The PHP Group                                |
   +----------------------------------------------------------------------+
   | This source file is subject to version 2.02 of the PHP license,      |
   | that is bundled with this package in the file LICENSE, and is        |
   | available at through the world-wide-web at                           |
   | http://www.php.net/license/2_02.txt.                                 |
   | If you did not receive a copy of the PHP license and are unable to   |
   | obtain it through the world-wide-web, please send a note to          |
   | license@php.net so we can mail you a copy immediately.               |
   +----------------------------------------------------------------------+
   | Author: David Eriksson <eriksson@php.net>                            |
   +----------------------------------------------------------------------+
 */

// $Id: object.cpp,v 1.15 2001/08/06 14:08:58 david Exp $
#include "object.h"
#include "zppclassentry.h"
#include "zppproxy.h"
#include "fromcorba.h"
#include "zppexception.h"
#include "zpp2corba.h"
#include "zppstream.h"

#include <memory>
using namespace std;

#if 0
static void IorProxy(INTERNAL_FUNCTION_PARAMETERS)
{
	ZppProxy::CallStaticFunction(&UniverseObject::Ior, INTERNAL_FUNCTION_PARAM_PASSTHRU);
}
#endif

const char * const gIorPropertyKey = "IOR";
const char * const gGetPrefix = "_get_";
const char * const gSetPrefix = "_get_";


ZppClassEntry * ZppClassStatic<UniverseObject>::mpClassEntry = NULL;
#if ENABLE_SATELLITE_COMPATIBLITY
const char * const ZppClassStatic<UniverseObject>::mpClassAlias = "OrbitObject";
#endif

zend_function_entry UniverseObject::mStaticFunctions[] =
{
	{"__sleep", ZppProxy::Sleep, NULL},
	{"__wakeup", ZppProxy::Wakeup, NULL},
#if 0
	{"ior", IorProxy, NULL},
#endif
	{NULL, NULL}
};


bool UniverseObject::RegisterClass()
{
	mpClassEntry = new ZppClassEntry(
		"UniverseObject",
		FactoryFunction,
		ZppClassEntry::hasEverything,
		mStaticFunctions);
	
#if ENABLE_SATELLITE_COMPATIBLITY
	new ZppClassEntry(
		mpClassAlias,
		mpClassEntry->FactoryFunction(),
		ZppClassEntry::hasEverything,		// mpClassEntry->...
		mStaticFunctions);							// mpClassEntry->...
#endif

	return true;
}

void UniverseObject::Create(ZppValueArray & rParameters)
{
	if (rParameters.Size() == 1)
	{
		// assuming IOR or path to IOR
		Create(rParameters[0].String());
	}
	else if (rParameters.Size() == 2)
	{
		// assuming object and repository id
		Create(rParameters[0], rParameters[1].String());
	}
	else
	{
		throw ZppWrongParameterCount_();
	}

#if 0
	Assert (rParameters.Size() == 1, ZppWrongParameterCount_());

	if (rParameters[0].IsObject())
	{
		// could be a UniverseImplementation
		UniverseImplementation* p_implementation = 
			UniverseImplementation::FindInstance(rParameters[0]);
		
		Create(p_implementation->Object());
	}
	else
	{
	}
#endif
}

void UniverseObject::Create(ZppValue& rObject, const char * pRepositoryId)
{
	try
	{
		// verify we have an object
		mObject = rObject;
		Assert(mObject.IsObject(), ZppTypeMismatchException_());
		
		// validate interface
		mInterface = CORBA::InterfaceDef::_narrow(
			Repository()->lookup_id(pRepositoryId));
		Assert(mInterface, ZppError_("Could not find interface"));

		// create CORBA object
		UniverseImplementation * p_implementation = new UniverseImplementation(
				mInterface->id(),
				mInterface->name(),
				this);

		mCorbaObject = p_implementation->_this();

		CORBA::Object_var poaobj = Orb()->resolve_initial_references ("RootPOA");
		PortableServer::POA_var poa = PortableServer::POA::_narrow (poaobj);
		PortableServer::POAManager_var mgr = poa->the_POAManager();

		mgr->activate ();
	}
	CATCH_CORBA_EXCEPTIONS();
}

void UniverseObject::Create(const char * pIor)
{
	try
	{
		Create(Orb()->string_to_object(pIor));
	}
	CATCH_CORBA_EXCEPTIONS();
}

void UniverseObject::Create(CORBA::Object_ptr pObject)
{
	try
	{
		mCorbaObject = pObject;
		
		const char * p_id = mCorbaObject->_repoid();

		mInterface = CORBA::InterfaceDef::_narrow(
			Repository()->lookup_id(p_id));
		Assert(mInterface, ZppError_("Could not find interface"));

	}
	CATCH_CORBA_EXCEPTIONS();
}

CORBA::OperationDef_ptr UniverseObject::FindOperation(
		CORBA::InterfaceDef_ptr pInterface, const char * pOperationName)
{
	// XXX: This is a depth-first search, maybe rewrite to breadth-first search?
	
	CORBA::OperationDef_ptr p_operation = 
		CORBA::OperationDef::_narrow(pInterface->lookup(pOperationName));
	
	if (p_operation)
		return p_operation;
	
	CORBA::InterfaceDefSeq* p_base_interfaces = pInterface->base_interfaces();

	if (p_base_interfaces)
	{
		// loop until all parents are searched or an operation is found
		for (unsigned i = 0; i < p_base_interfaces->length() && !p_operation; i++)
		{
			p_operation = FindOperation( (*p_base_interfaces)[i], pOperationName );
		}
	}

	return p_operation;
}


/**
 * Find operation info
 */
CORBA::OperationDef_ptr UniverseObject::FindOperation(const char * pOperationName)
{
	// Prevent crash
	Assert(mInterface, ZppNullPointerException_());

#if 0
	OperationDef_map::iterator item = mOperationDefCache.find(pOperationName);
	if (mOperationDefCache.end() == item)
	{
#endif
		CORBA::OperationDef_ptr p_operation = 
			FindOperation(mInterface, pOperationName);
		Assert(p_operation, ZppError_("Could not find operation"));
//		mOperationDefCache[pOperationName] = p_operation;
		return p_operation;
#if 0
	}
	else
	{
		item->second->_ref();
		return item->second;
	}
#endif
}

/**
 * Create exception list for operation
 */
void UniverseObject::CreateExceptionList(CORBA::OperationDef_ptr operation, 
		CORBA::ExceptionList_out exceptionList)
{
	Orb()->create_exception_list(exceptionList);

	for (unsigned i = 0; i < operation->exceptions()->length(); i++)
	{
		exceptionList->add( (*operation->exceptions())[i]->type() );
	}
}

/**
 * Create named value for result
 */
void UniverseObject::CreateResult(CORBA::OperationDef_ptr operation,
		CORBA::NamedValue_out result)
{
	CORBA::TypeCode_var result_type(operation->result());
	if (result_type)
	{
		Orb()->create_named_value(result);
		result->value()->set_type(result_type);
	}
}

/**
 * Create argument list
 */
void UniverseObject::CreateArgumentList(CORBA::OperationDef_ptr operation,
		ZppValueArray & rParameters, CORBA::NVList_var & rArguments)
{
	Assert(rParameters.Size() == operation->params()->length(),
			ZppWrongParameterCount_());

	// This will create a list with flags and types set correctly
	Orb()->create_operation_list(operation, rArguments);

	Assert(rParameters.Size() == rArguments->count(),
			ZppWrongParameterCount_());

	for (unsigned i = 0; i < rParameters.Size(); i++)
	{
		CORBA::NamedValue_ptr namedvalue = rArguments->item(i);
		
		// Convert input parameters only
		CORBA::Flags flags = namedvalue->flags();
		if ( (flags & CORBA::ARG_IN) || (flags & CORBA::ARG_INOUT) )
		{
			zpp2corba(rParameters[i], *namedvalue->value());
		}
	}
}

void UniverseObject::ConvertOutputParameters(CORBA::NVList_ptr arguments, 
		ZppValueArray & rParameters)
{
	for (unsigned i = 0; i < rParameters.Size(); i++)
	{
		CORBA::NamedValue_ptr namedvalue = arguments->item(i);
		
		// Convert output parameters only
		CORBA::Flags flags = namedvalue->flags();
		if ( (flags & CORBA::ARG_OUT) || (flags & CORBA::ARG_INOUT) )
		{
			from_corba(namedvalue, rParameters[i]);
		}
	}
}

void UniverseObject::CallFunction(const char * pFunctionName, 
	ZppValueArray & rParameters, ZppValue & rReturnValue)
{
	try 
	{
#if 0
		if (strcasecmp(pFunctionName, "ior") == 0)
		{
			Assert(rParameters.Size() == 0, ZppWrongParameterCount_());
			char* p_ior = Orb()->object_to_string(mCorbaObject);
			Assert(p_ior, ZppNullPointerException_());
			rReturnValue.String(p_ior);
			return;	
		}
#endif
		
		CORBA::OperationDef_var operation( FindOperation(pFunctionName) );
		
		//
		// Get exception list from operation info
		//
		CORBA::ExceptionList_var exception_list;
		CreateExceptionList(operation, exception_list);
			
		//
		// Create result value
		//
		CORBA::NamedValue_var result;
		CreateResult(operation, result);

		//
		// Add arguments
		//
		CORBA::NVList_var arguments;
		CreateArgumentList(operation, rParameters, arguments);

		//
		// Create request
		//
		CORBA::Request_var request;

		mCorbaObject->_create_request(
				NULL,	/* context */
				pFunctionName,
				arguments, /* argument list */
				result,
				exception_list,
				NULL, /* context list */
				request,
				0 		/* flags */	
				);
				
		
		//
		// Perform request
		//
		request->invoke();
		
		//
		// TODO: Take care of output arguments
		//
		ConvertOutputParameters(arguments, rParameters);
		
		//
		// Take care of return value
		//
		if (result)
		{
//			cerr << "[CallFunction] Result length = " << result->value()->length() << endl;
			from_corba(result, rReturnValue);
		}
		else
		{
			rReturnValue.Null();
		}

	}
	CATCH_CORBA_EXCEPTIONS();
}

CORBA::AttributeDef_ptr UniverseObject::FindAttribute(const char * pAttributeName)
{
	CORBA::AttributeDef_ptr p_attribute = 
		CORBA::AttributeDef::_narrow(mInterface->lookup(pAttributeName));
	Assert(p_attribute, ZppError_("Could not find attribute"));
	return p_attribute;

}

void UniverseObject::GetProperty(const char * pPropertyName, ZppValue& rValue)
{
	try	
	{
		//
		// Get attribute info
		//
		CORBA::AttributeDef_var attribute( FindAttribute(pPropertyName) );

		//
		// create name for request
		//
		strstream_zpp name;
		name << gGetPrefix << pPropertyName << ends;

		//
		// create result value
		//
		CORBA::NamedValue_var result;
		Orb()->create_named_value(result);
		result->value()->set_type(attribute->type());

		//
		// Create request
		//
		CORBA::Request_var request;

		mCorbaObject->_create_request(
				NULL,	/* context */
				name.str(),
				NULL, /* argument list */
				result,
				request,
				0 		/* flags */	
				);

		//
		// Perform request
		//
		request->invoke();
		
		//
		// Handle return value
		//
		from_corba(result, rValue);
	}
	CATCH_CORBA_EXCEPTIONS();
}

void UniverseObject::PutProperty(const char * pPropertyName, ZppValue& rValue)
{
	try
	{
		//
		// Get attribute info
		//
		CORBA::AttributeDef_var attribute( FindAttribute(pPropertyName) );

		// Verify it's a writable attribute
		if (attribute->mode() == CORBA::ATTR_READONLY)
		{
			throw ZppError_("Property is read-only");
		}
		
		//
		// create name for request
		//
		strstream_zpp name;
		name << gSetPrefix << pPropertyName << ends;
	
		//
		// create argument list with one element
		// 
		CORBA::NVList_var arguments;
		Orb()->create_list(0, arguments);		// XXX: what's the first param?

		//
		// Add a namedvalue to list and set value
		//
		CORBA::NamedValue_ptr namedvalue = arguments->add(CORBA::ARG_IN);
		zpp2corba(rValue, *namedvalue->value());

		//
		// Create request
		//
		CORBA::Request_var request;

		mCorbaObject->_create_request(
				NULL,	/* context */
				name.str(),
				arguments, /* argument list */
				NULL, /* result */
				request,
				0 		/* flags */	
				);

		//
		// Perform request
		//
		request->invoke();
	}
	CATCH_CORBA_EXCEPTIONS();
}

void UniverseObject::Sleep(ZppValue& rObject, ZppValue& rPropertyNames)
{
	// get IOR from object
	char* p_ior = Orb()->object_to_string(mCorbaObject);
	Assert(p_ior, ZppNullPointerException_());
		
	// add property to our object
	rObject.AddPropertyString(gIorPropertyKey, p_ior);

	// add property name to array
	rPropertyNames.AddNextIndexString(gIorPropertyKey);

	// XXX: leaking p_ior?
}

void UniverseObject::Wakeup(ZppValue& rObject)
{
#if 0
	// get IOR
	zval ** pp_ior = NULL;
	bool success = Value().Properties().Find(
			gIorPropertyKey,	// key
			&pp_ior						// pointer heaven
			);

	// verify things
	Assert(success && pp_ior && *pp_ior, ZppNullPointerException_());
#endif
	// call Create method with IOR
	ZppValue ior( rObject.Properties()[gIorPropertyKey] );
	Create(ior.String());
}

void UniverseObject::ConvertFromCorba(CORBA::Any& rSource, ZppValue& rResult, 
		CORBA::TypeCode_ptr pType)
{
	CORBA::Object_var value;
	rSource >>= CORBA::Any::to_object(value);
	
	UniverseObject * p_object = new UniverseObject;
	p_object->Create(CORBA::Object::_duplicate(value));
	
	rResult.Object(p_object);
}
		
void UniverseObject::ConvertToCorba(ZppValue& rSource, CORBA::Any& rDestination, 
		CORBA::TypeCode_ptr pType)
{
	// XXX: mico extension
	rDestination <<= CORBA::Any::from_object( 
			FindInstance(rSource)->mCorbaObject, pType->name() );
}

#if 0
void UniverseObject::Ior(ZppValueArray & rParameters, ZppValue& rReturnValue)
{
}
#endif

void UniverseObject::Invoke (CORBA::ServerRequest_ptr request)
{
	// ZEND_API int call_user_function(HashTable *function_table, 
	// 		zval **object_pp, zval *function_name, zval *retval_ptr, 
	//		int param_count, zval *params[]);
	// ZEND_API int call_user_function_ex(HashTable *function_table, 
	//		zval **object_pp, zval *function_name, zval **retval_ptr_ptr, 
	//		int param_count, zval **params[], int no_separation, HashTable *symbol_table);

	//
	// Get operation info
	//
	CORBA::OperationDef_var operation( /*request->op_def()*/ FindOperation(request->op_name()) );

	// MEMORY: ServerRequest takes ownership of the params object
	//         (see comments in include/mico/dsi.h)
	CORBA::NVList_ptr params = NULL;
	Orb()->create_operation_list(operation, params);
	CORBA::Boolean success = request->params(params);
	Assert( success, ZppError_("Call to ServerRequest::params() failed") );

	//
	// Prepare ZppValueArray for parameters
	//
	int parameter_count = operation->params()->length();
	
	// MEMORY: pp_parameters is freed by the auto_ptr
	zval ** pp_parameters = new zval*[parameter_count];
	auto_ptr<zval*> cleanup_parameters(pp_parameters);
	
	for (int i = 0; i < parameter_count; i++)
	{
		// MEMORY: the zval is freed by the ZppValueArray
		MAKE_STD_ZVAL(pp_parameters[i]);
		pp_parameters[i]->type = IS_NULL;
		// we want the ZppValue in ZppValueArray to take ownership of the zval
		pp_parameters[i]->refcount = 0;
	}
	ZppValueArray parameters(parameter_count, pp_parameters);
	
	//
	// Convert input parameters
	// NOTE: arguments() is a MICO extension
	//
	ConvertZendInputParameters(request->arguments(), parameters);

	//
	// Prepare some parameters to call_user_function()
	//
	zval * p_object = mObject.Get();
	ZppValue return_value; return_value.Null();
	ZppValue function_name( operation->name() );

	//
	// Make function call
	//
	int rv = ::call_user_function(NULL, &p_object, function_name.Get(), 
			return_value.Get(), parameter_count, pp_parameters);

	if (FAILURE == rv)
	{
		// TODO: set exception
		cerr << "[UniverseObject::Invoke] call_user_function failed" << endl;
		return;
	}

	//
	// Convert output parameters
	// NOTE: arguments() is a MICO extension
	//
	ConvertZendOutputParameters(parameters, request->arguments());

	//
	// Convert return value
	//
	if (operation->result() && operation->result()->kind() != CORBA::tk_void)
	{
		CORBA::Any result;
		result.set_type(operation->result());
		zpp2corba(return_value, result);
		request->set_result(result);
	}
}

void UniverseObject::ConvertZendInputParameters(CORBA::NVList_ptr source, 
		ZppValueArray & rDestination)
{
	Assert(
			(NULL == source && rDestination.Size() == 0) ||
			(NULL != source && rDestination.Size() == source->count()),
			ZppWrongParameterCount_());

	for (unsigned i = 0; i < rDestination.Size(); i++)
	{
		CORBA::NamedValue_ptr namedvalue = source->item(i);
		CORBA::Flags flags = namedvalue->flags();

		// Convert input parameters only
		if ( (flags & CORBA::ARG_IN) || (flags & CORBA::ARG_INOUT) )
		{
			from_corba(namedvalue, rDestination[i]);
		}

		// Make output values references
		if ( (flags & CORBA::ARG_OUT) || (flags & CORBA::ARG_INOUT) )
		{
			rDestination[i].SetReference(true);
		}
	}
}

void UniverseObject::ConvertZendOutputParameters(ZppValueArray & rSource, 
		CORBA::NVList_ptr pDestination)
{
	for (unsigned i = 0; i < rSource.Size(); i++)
	{
		CORBA::NamedValue_ptr namedvalue = pDestination->item(i);
		
		// Convert output parameters only
		CORBA::Flags flags = namedvalue->flags();
		if ( (flags & CORBA::ARG_OUT) || (flags & CORBA::ARG_INOUT) )
		{
			zpp2corba(rSource[i], *namedvalue->value());
		}
	}
}

char * UniverseObject::Ior()
{
	return Orb()->object_to_string(mCorbaObject);
}
