/*
   +----------------------------------------------------------------------+
   | PHP version 4.0                                                      |
   +----------------------------------------------------------------------+
   | Copyright (c) 1997-2001 The PHP Group                                |
   +----------------------------------------------------------------------+
   | This source file is subject to version 2.02 of the PHP license,      |
   | that is bundled with this package in the file LICENSE, and is        |
   | available at through the world-wide-web at                           |
   | http://www.php.net/license/2_02.txt.                                 |
   | If you did not receive a copy of the PHP license and are unable to   |
   | obtain it through the world-wide-web, please send a note to          |
   | license@php.net so we can mail you a copy immediately.               |
   +----------------------------------------------------------------------+
   | Author: David Eriksson <eriksson@php.net>                            |
   +----------------------------------------------------------------------+
 */

#include "universe.h"
//#include "zppregistry.h"
#include "object.h"
#include "struct.h"
#include "implementation.h"
#include "zppclassentry.h"

PHP_FUNCTION(universe_run)
{
	try
	{
		Assert(ZEND_NUM_ARGS() == 0, ZppWrongParameterCount_());
		
		try
		{
			UniverseContext::Get().Orb()->run();
		}
		CATCH_CORBA_EXCEPTIONS();
	}
	CATCH_COMMON_EXCEPTIONS();
}

PHP_FUNCTION(universe_object_to_string)
{
	try
	{
		zval **arg = NULL;
		Assert(ZEND_NUM_ARGS() == 1 && zend_get_parameters_ex(1, &arg) == SUCCESS, 
				ZppWrongParameterCount_());

		ZppValue object(*arg);

		UniverseObject * p_instance = UniverseObject::FindInstance(object);
		Assert(p_instance, ZppTypeMismatchException_());

		ZppValue result(return_value);
		
		try
		{
			result.String(p_instance->Ior());
		}
		CATCH_CORBA_EXCEPTIONS();
	}
	CATCH_COMMON_EXCEPTIONS();
}

#if 0
CORBA::ORB_var Universe::mOrb;
CORBA::Repository_var Universe::mRepository;
#endif

zend_bool universe_initialize()
{
	try
	{
		ZppModule::Set( new UniverseModule() );
		ZppModule::Get().Initialize();
		return true;
	}
	catch (...)
	{
		// TODO: handle error
		return false;
	}
}

void universe_shutdown()
{
	ZppModule::Destroy();
}
		
UniverseContext::~UniverseContext()
{
#if 0
	if (mOrb)
		DUMP(mOrb->_refcnt());
	if (mRepository)
		DUMP(mRepository->_refcnt());
#endif
}

UniverseContext& UniverseContext::Get()
{
	return UniverseModule::Get().ActiveContext();
}

void UniverseModule::Initialize()
{
	ZppModule::Initialize();		// Very important to call parent class here

	UniverseObject::RegisterClass();
	UniverseStruct::RegisterClass();
//	ZppRegistry::Get().RegisterClassEntries();
}

bool UniverseContext::InitializeCorba()
{
	int argc = 5;
	char * argv[] = { "universe", 
										 "-ORBDebugLevel", INI_STR(INI_ORBDebugLevel),
										 "-ORBIfaceRepoAddr", INI_STR(INI_ORBIfaceRepoAddr),
										 NULL, NULL,	// for ORBIIOPAddr
										 NULL };

	try
	{
		// Maybe we have the ORBIIOPAddr option too?
		char * p_ORBIIOPAddr = INI_STR(INI_ORBIIOPAddr);
		if (p_ORBIIOPAddr && *p_ORBIIOPAddr)
		{
			argv[argc++] = "-ORBIIOPAddr";
			argv[argc++] = p_ORBIIOPAddr;
		}

		mOrb = CORBA::ORB_init(argc, argv, "mico-local-orb");
		
		 CORBA::Object_var obj =
      mOrb->resolve_initial_references ("InterfaceRepository");
    mRepository = CORBA::Repository::_narrow (obj);

		return true;
	}
/*	catch (CORBA::SystemException & e)
	{
	}
	catch (CORBA::Exception & e)
	{
	}*/
	catch (...)
	{
		fprintf(stderr, "Universe::InitializeCorba - exception\n");
	}

	return false;
}

//! get pointer to ORB
CORBA::ORB_ptr UniverseContext::Orb()
{
	if (!mOrb)
	{
		InitializeCorba();
	}
	return mOrb;
}

//! get pointer to interface repository
CORBA::Repository_ptr UniverseContext::Repository()
{
	if (!mRepository)
	{
		InitializeCorba();
	}
	return mRepository;
}


